/*
    psftools: Manipulate console fonts in the .PSF format
    Copyright (C) 2000, 2021  John Elliott

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include "cnvshell.h"
#include "psflib.h"
#include "mswfnt.h"

#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif

/* Convert a Windows .FNT font to Berkeley vfont
 * Note that this is not the same as the .FON fonts distributed with 
 * MS-Windows. You have to extract the .FNT fonts from the .FON using 
 * a resource editor. */

static char helpbuf[2048];

static VFONT vfont;
static PSF_MAPPING *codepage = NULL;
static MSW_FONTINFO msw;
static int v1 = 0, v2 = 0;
static int basepoint_x, have_basepoint;
static int big_endian = 0;
/* Program name */
char *cnv_progname = "FNT2VFONT";

/* ddash = 1 if option started with a double-dash; else 0 */
/* Return NULL if OK, else error string */
char *cnv_set_option(int ddash, char *variable, char *value)
{
	if (!stricmp(variable, "v1"))   { v1 = 1; return NULL; }
	if (!stricmp(variable, "v2"))   { v2 = 1; return NULL; }
	if (!stricmp(variable, "be") || !stricmp(variable, "big-endian")) 
		{ big_endian = 1; return NULL; }
	if (!stricmp(variable, "le") || !stricmp(variable, "little-endian")) 
		{ big_endian = 0; return NULL; }
	if (!stricmp(variable, "basepoint"))
	{
		if (value == NULL || sscanf(value, "%d", &basepoint_x) != 1)
		{
			return "The -basepoint option must be followed by a number.";
		}
		have_basepoint = 1;	
		return NULL;
	}

	if (!stricmp(variable, "codepage"))
	{
		codepage = psf_find_mapping(value);
		if (codepage == NULL) return "Code page name not recognised.";
		return NULL;
	}
	if (strlen(variable) > 2000) variable[2000] = 0;
	sprintf(helpbuf, "Unknown option: %s\n", variable);
	return helpbuf;
}


/* Return help string */
char *cnv_help(void)
{
	sprintf(helpbuf, "Syntax: %s winfnt psffile { options }\n\n", 
			cnv_progname);
	strcat (helpbuf, "Options: \n"
	                 "    --v1            - Output in original vfont format (default)\n"
	                 "    --v2            - Output in proposed vfont2 format\n"
			 "    --big-endian    - Output a big-endian file (v1 only).\n"
			 "    --basepoint=x   - Set left basepoint of characters (default is the\n"
			 "                      left edge of capital A)\n"
	                 "    --codepage=x    - Create a Unicode directory "
			 "from the specified code page\n");
	return helpbuf;
}


static msw_word offset_table[768];
static psf_dword stdpos;

static int maybe_fseek(FILE *fp, long pos)
{
	if (fp != stdin) return fseek(fp, pos, SEEK_SET);

	while (pos > stdpos) { getchar(); ++stdpos; }
	return 0;
}

/* Convert Windows 1.x font */
char *msw_conv_1(FILE *fp)
{
	int y, x, vx, x1, ch, chx;
	psf_byte b, mask;
	psf_dword bitmap = 0;

	stdpos = 0x75;

	/* v1 file only has a lengths table for proportional fonts, and all 
         * the characters are stored side-by-side in a bitmap. */
	if (msw.dfPixWidth == 0) 
	{
		for (ch = msw.dfFirstChar; ch <= msw.dfLastChar+1; ch++)
		{
			if (read_word(fp, &offset_table[ch])) 
			{
				return "Unexpected end of file";
			}
			stdpos += 2;
		}
	}
	else
	{
		vx = 0;
		for (ch = msw.dfFirstChar; ch <= msw.dfLastChar+1; ch++)
		{
			offset_table[ch] = vx;
			vx += msw.dfPixWidth;
		}
	}
	/* Set character dimensions for all characters */
	for (ch = msw.dfFirstChar; ch <= msw.dfLastChar; ch++) 
	{
		vfont.vf_dispatch[ch].vfd_up    = msw.dfAscent;
		vfont.vf_dispatch[ch].vfd_down  = msw.dfPixHeight - vfont.vf_dispatch[ch].vfd_up;
		if (msw.dfPixWidth == 0) /* Variable */
		{
			vfont.vf_dispatch[ch].vfd_left  = 0;
			vfont.vf_dispatch[ch].vfd_right = offset_table[ch + 1] - offset_table[ch];
			vfont.vf_dispatch[ch].vfd_nbytes = 
				((vfont.vf_dispatch[ch].vfd_right + 7) / 8) * msw.dfPixHeight;
		}
		else			/* Fixed */
		{
			vfont.vf_dispatch[ch].vfd_left  = 0;
			vfont.vf_dispatch[ch].vfd_right = msw.dfPixWidth;
			vfont.vf_dispatch[ch].vfd_nbytes = 
				((msw.dfPixWidth + 7) / 8) * msw.dfPixHeight;

		}
		vfont.vf_dispatch[ch].vfd_width = vfont.vf_dispatch[ch].vfd_left + 
							vfont.vf_dispatch[ch].vfd_right;
		vfont.vf_dispatch[ch].vfd_addr = bitmap;
		bitmap += vfont.vf_dispatch[ch].vfd_nbytes;
	}


	/* Seek to font bitmap */
	if (maybe_fseek(fp, msw.dfBitsOffset)) 
		return "Unexpected end of file";	
	for (y = 0; y < msw.dfPixHeight; y++)
	{
		for (x = 0; x < msw.dfWidthBytes; x++)
		{
			if (read_byte(fp, &b)) return "Unexpected end of file";
			vx = x * 8;
			mask = 0x80;
			for (x1 = 0; x1 < 8; x1++,vx++)
			{
/* Find out in which character the byte lies */
				for (ch = msw.dfFirstChar; ch <= msw.dfLastChar; ch++) 
					if (offset_table[ch] <= vx &&
					    offset_table[ch+1] > vx) break;

				if (ch > msw.dfLastChar) break;	/* Off the edge of the bitmap */
				chx = vx - offset_table[ch];

				vfont_set_pixel(&vfont, ch, chx, y, b & mask);	
				mask = mask >> 1;
			}
		}
	}
	return NULL;
}

/* Convert Windows 2.x font */

char *msw_conv_2(FILE *fp)
{
	int ch, y, ww, x;
	psf_byte *b;
	int c;
	psf_dword bitmap = 0;

	stdpos = 0x76;
	for (ch = msw.dfFirstChar; ch <= msw.dfLastChar; ch++)
	{
		if (read_word(fp, &offset_table[ch*2])   ||
		    read_word(fp, &offset_table[ch*2+1])) 
			return "Unexpected end of file";
		stdpos += 4;
	}
	/* Set character dimensions for all characters */
	for (ch = msw.dfFirstChar; ch <= msw.dfLastChar; ch++) 
	{
		vfont.vf_dispatch[ch].vfd_up    = msw.dfAscent;
		vfont.vf_dispatch[ch].vfd_down  = msw.dfPixHeight - vfont.vf_dispatch[ch].vfd_up;
		vfont.vf_dispatch[ch].vfd_left  = 0;
		vfont.vf_dispatch[ch].vfd_right = offset_table[ch * 2];
		vfont.vf_dispatch[ch].vfd_nbytes = 
				((vfont.vf_dispatch[ch].vfd_right + 7) / 8) * msw.dfPixHeight;
		vfont.vf_dispatch[ch].vfd_width = vfont.vf_dispatch[ch].vfd_left + 
							vfont.vf_dispatch[ch].vfd_right;
		vfont.vf_dispatch[ch].vfd_addr = bitmap;
		bitmap += vfont.vf_dispatch[ch].vfd_nbytes;
	}


	for (ch = msw.dfFirstChar; ch <= msw.dfLastChar; ch++)
	{
		ww = (offset_table[ch*2] + 7) / 8;
		maybe_fseek(fp, offset_table[ch*2+1]);
		b = &vfont.vf_bitmap[vfont.vf_dispatch[ch].vfd_addr];
		for (x = 0; x < ww; x++)
		{
			for (y = 0; y < msw.dfPixHeight; y++)
			{
				c = fgetc(fp);	
				++stdpos;
				if (c == EOF) return "Unexpected end of file";
				b[y * ww + x] = c;
			}
		}
	}
	return NULL;
}

/* Convert Windows 3.x font */

char *msw_conv_3(FILE *fp)
{
	int ch, y, ww, x;
	psf_byte *b;
	int c;
	long offset;
	psf_dword bitmap = 0;

	stdpos = 0x9A;
        for (ch = msw.dfFirstChar; ch <= msw.dfLastChar; ch++)
        {
                if (read_word(fp, &offset_table[ch*3])   ||
                    read_word(fp, &offset_table[ch*3+1]) ||
		    read_word(fp, &offset_table[ch*3+2]))
                        return "Unexpected end of file";
		stdpos += 6;
        }
	/* Set character dimensions for all characters */
	for (ch = msw.dfFirstChar; ch <= msw.dfLastChar; ch++) 
	{
		vfont.vf_dispatch[ch].vfd_up    = msw.dfAscent;
		vfont.vf_dispatch[ch].vfd_down  = msw.dfPixHeight - vfont.vf_dispatch[ch].vfd_up;
		vfont.vf_dispatch[ch].vfd_left  = 0;
		vfont.vf_dispatch[ch].vfd_right = offset_table[ch * 3];
		vfont.vf_dispatch[ch].vfd_nbytes = 
				((vfont.vf_dispatch[ch].vfd_right + 7) / 8) * msw.dfPixHeight;
		vfont.vf_dispatch[ch].vfd_width = vfont.vf_dispatch[ch].vfd_left + 
							vfont.vf_dispatch[ch].vfd_right;
		vfont.vf_dispatch[ch].vfd_addr = bitmap;
		bitmap += vfont.vf_dispatch[ch].vfd_nbytes;
	}
        for (ch = msw.dfFirstChar; ch <= msw.dfLastChar; ch++)
        {
		ww = (offset_table[ch*3] + 7) / 8;
		offset = offset_table[ch*3+2];
		offset = (offset << 16) | offset_table[ch*3+1];
                maybe_fseek(fp, offset);
		b = &vfont.vf_bitmap[vfont.vf_dispatch[ch].vfd_addr];
		for (x = 0; x < ww; x++)
		{
			for (y = 0; y < msw.dfPixHeight; y++)
			{
				c = fgetc(fp);	
				++stdpos;
				if (c == EOF) return "Unexpected end of file";
				b[y * ww + x] = c;
			}
		}
        }
        return NULL;
}




/* Do the conversion */
char *cnv_execute(FILE *infile, FILE *outfile)
{	
	int rv;
	char *rvs;
	psf_dword bitmap_size;
	int version;

	if (v2) version = 2;
	else version = 1;
	if (big_endian && version == 2)
	{
		return "A version 2 vfont cannot be big-endian";
	}

	vfont_new(&vfont);

	rv = msw_fontinfo_read(&msw, infile);

	if (rv == -1) return "Read error on input file";
	if (rv == -2) return ".FNT file is not in a recognised Windows format";

	if (msw.dfType & 0x7F) return ".FNT file is not a bitmap font";

	bitmap_size = ((msw.dfMaxWidth + 7) / 8) * msw.dfPixHeight * (msw.dfLastChar + 1);

	rv = vfont_create(&vfont, msw.dfLastChar + 1, bitmap_size, 0);
	if (rv) return psf_error_string(rv);
	
	rvs = "Unsupported Windows font format";
	switch(msw.dfVersion)
	{
		case 0x100: rvs = msw_conv_1(infile); break;
		case 0x200: rvs = msw_conv_2(infile); break;
		case 0x300: rvs = msw_conv_3(infile); break;
	}
	if (!codepage)
	{
		if (msw.dfCharSet == 0xFF)   codepage = psf_find_mapping("CP437");
		else if (msw.dfCharSet == 0) codepage = psf_find_mapping("CP1252");
	}
        if (codepage && v2)
        {
		int n;
                vfont_create_unicode(&vfont);
                for (n = msw.dfFirstChar; n < msw.dfLastChar; n++) if (n < 256)
		{
                	vfont_unicode_addmap(&vfont, n, codepage, n);
		}
	}
/* Find the basepoint */
	if (!have_basepoint && vfont.vf_length >= 'A')
	{
		int y, x, w;
		psf_byte pixel;		

		w = vfont.vf_dispatch['A'].vfd_left + vfont.vf_dispatch['A'].vfd_right;	
		for (x = 0; x < w; x++)
		{
			for (y = msw.dfPixHeight - 1; y >= 0; y--)		
			{
				vfont_get_pixel(&vfont, 'A', x, y, &pixel);
				if (!vfont_get_pixel(&vfont, 'A', x, y, &pixel) && pixel != 0)
				{	
					basepoint_x = x;
					have_basepoint = 1;
					goto found;
				}
			}
		}
		found: ;	/* Break out of double loop */
		fprintf(stderr, "Basepoint X set to %d\n", basepoint_x);
	}
	/* Apply basepoint_x to all characters */
	if (have_basepoint)
	{
		int x;

		for (x = 0; x < vfont.vf_length; x++)
		{
			vfont.vf_dispatch[x].vfd_left   = basepoint_x;
			vfont.vf_dispatch[x].vfd_right -= basepoint_x;
		}
	}


        if (v1) vfont_force_v1(&vfont);
        if (v2) vfont_force_v2(&vfont);

	if (!rvs) rv = vfont_write(&vfont, outfile);
	vfont_delete(&vfont);	
	if (rvs) return rvs;
	if (rv) return psf_error_string(rv);
	return NULL;
}

